const jwt = require('jsonwebtoken');
const { verifyAwsToken } = require('./verifyAwsToken');

const TOKEN_TYPE = {
    COGNITO: 'cognito',
    LOCAL: 'local',
};

/**
 * @param token
 * @return {TOKEN_TYPE} [cognito|local]
 */
function getTokenIssuer(token) {
    /** @type {AwsAccessTokenType|RppTokenType|CompleteTokenType|string|null} */
    const decoded = jwt.decode(token, { complete: false });
    const issuer = decoded.iss;
    return issuer && issuer.includes('cognito-idp') && issuer.includes('amazonaws.com')
        ? TOKEN_TYPE.COGNITO
        : TOKEN_TYPE.LOCAL;
}

function validateCognitoToken(token) {
    return new Promise((resolve, reject) => {
        verifyAwsToken(token)
            .then((decodedToken) => {
                if ((decodedToken.iat + decodedToken.exp) * 1000 > new Date().getTime()) {
                    return resolve({ error: null });
                }
                // eslint-disable-next-line prefer-promise-reject-errors
                return reject({ error: true, err: 'AWS token expired' });
            })
            .catch((err) => reject(Error({ error: true, err })));
    });
}

function validateLocalToken(token) {
    return new Promise((resolve, reject) => {
        jwt.verify(
            token,
            process.env.RPP_GW_AUTH_TOKEN_SECRET,
            /**
             * @param err
             * @param {AwsAccessTokenType| RppTokenType} decodedToken
             */
            (err, decodedToken) => {
                if (!err && (decodedToken.iat + (decodedToken.expiresIn || decodedToken.exp)) * 1000 > new Date().getTime()) {
                    return resolve({ error: null });
                }
                // eslint-disable-next-line prefer-promise-reject-errors
                return reject({ error: true, err });
            },
        );
    });
}

function validateToken(token) {
    const issuer = getTokenIssuer(token);

    if (issuer === TOKEN_TYPE.COGNITO) {
        return validateCognitoToken(token);
    }

    return validateLocalToken(token);
}

module.exports = {
    validateToken,
};
